from typing import Dict, Optional, Iterable, Sequence, Type, Tuple, TypeVar, Generic
from pydantic import BaseModel
from .pagination import Edge, encode_pagination_cursor, decode_pagination_cursor, get_pagination_cursor_payload

T = TypeVar('T', bound=BaseModel)
OutputT = TypeVar('OutputT', bound=BaseModel)

Sort = Sequence[Tuple[str, int]]


class AbstractRepository(Generic[T]):
    def __init__(self, database):
        super().__init__()
        self.__database = database
        self.__document_class = self.__orig_bases__[0].__args__[0]
        self.__collection_name = self.Meta.collection_name
        self.__validate()

    def get_collection(self):
        return self.__database[self.__collection_name]

    def __validate(self):
        if not issubclass(self.__document_class, BaseModel):
            raise Exception('Document class should inherit BaseModel')
        if 'id' not in self.__document_class.__fields__:
            raise Exception('Document class should have id field')
        if not self.__collection_name:
            raise Exception('Meta should contain collection name')

    def to_document(self, model: T) -> dict:
        result = model.dict()
        result.pop('id')
        if model.id:
            result['_id'] = model.id
        return result

    def __map_id(self, data: dict) -> dict:
        query = data.copy()
        if 'id' in data:
            query['_id'] = query.pop('id')
        return query

    def __map_sort(self, sort: Sort) -> Optional[Sort]:
        result = []
        for item in sort:
            key = item[0]
            ordering = item[1]
            if key == 'id':
                key = '_id'
            result.append((key, ordering))
        return result

    def to_model_custom(self, output_type: Type[OutputT], data: dict) -> OutputT:
        data_copy = data.copy()
        if '_id' in data_copy:
            data_copy['id'] = data_copy.pop('_id')
        return output_type.parse_obj(data_copy)

    def to_model(self, data: dict) -> T:
        return self.to_model_custom(self.__document_class, data)

    def save(self, model: T):
        document = self.to_document(model)

        if model.id:
            mongo_id = document.pop('_id')
            self.get_collection().update_one({'_id': mongo_id}, {'$set': document})
            return

        result = self.get_collection().insert_one(document)
        model.id = result.inserted_id
        return result

    def delete(self, model: T):
        return self.get_collection().delete_one({'_id': model.id})

    def find_one_by_id(self, id: str) -> Optional[T]:
        return self.find_one_by({'id': id})

    def find_one_by(self, query: dict) -> Optional[T]:
        result = self.get_collection().find_one(self.__map_id(query))
        return self.to_model(result) if result else None

    def find_by_with_output_type(
        self,
        output_type: Type[OutputT],
        query: dict,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None
    ) -> Iterable[OutputT]:
        mapped_projection = self.__map_id(projection) if projection else None
        mapped_sort = self.__map_sort(sort) if sort else None
        cursor = self.get_collection().find(self.__map_id(query), mapped_projection)
        if limit:
            cursor.limit(limit)
        if skip:
            cursor.skip(skip)
        if sort:
            cursor.sort(mapped_sort)
        return map(lambda doc: self.to_model_custom(output_type, doc), cursor)

    def find_by(
        self,
        query: dict,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None
    ) -> Iterable[T]:
        return self.find_by_with_output_type(
            output_type=self.__document_class,
            query=query,
            skip=skip,
            limit=limit,
            sort=sort,
            projection=projection
        )

    def get_pagination_query(
        self,
        query: dict,
        after: Optional[str] = None,
        before: Optional[str] = None,
        sort: Optional[Sort] = None
    ) -> dict:
        generated_query: dict = {'$and': [query]}
        selected_cursor = after or before

        if selected_cursor and sort:
            cursor_data = decode_pagination_cursor(selected_cursor)
            dict_values = []
            for i, sort_expression in enumerate(sort):
                if after:
                    compare_operator = '$gt' if sort_expression[1] > 0 else '$lt'
                else:
                    compare_operator = '$lt' if sort_expression[1] > 0 else '$gt'
                dict_values.append((
                    sort_expression[0],
                    {compare_operator: cursor_data[i]}
                ))
            generated_query['$and'].append(dict(dict_values))

        if len(generated_query['$and']) == 1:
            generated_query = query or {}

        return generated_query

    def paginate_with_output_type(
        self,
        output_type: Type[OutputT],
        query: dict,
        limit: int,
        after: Optional[str] = None,
        before: Optional[str] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None
    ) -> Iterable[Edge[OutputT]]:
        sort_keys = []

        if not sort:
            sort = [('_id', 1)]

        for sort_expression in sort:
            sort_keys.append(sort_expression[0])

        models = self.find_by_with_output_type(
            output_type,
            query=self.get_pagination_query(
                query=query,
                after=after,
                before=before,
                sort=sort
            ),
            limit=limit,
            sort=sort,
            projection=projection
        )

        return map(
            lambda model: Edge[T](
                node=model,
                cursor=encode_pagination_cursor(get_pagination_cursor_payload(model, sort_keys))
            ),
            models
        )

    def paginate(
        self,
        query: dict,
        limit: int,
        after: Optional[str] = None,
        before: Optional[str] = None,
        sort: Optional[Sort] = None,
        projection: Optional[Dict[str, int]] = None
    ) -> Iterable[Edge[T]]:
        return self.paginate_with_output_type(
            self.__document_class,
            query,
            limit,
            after=after,
            before=before,
            sort=sort,
            projection=projection
        )



import datetime
from bson import ObjectId
from pydantic_mongo.pagination import encode_pagination_cursor, decode_pagination_cursor, get_pagination_cursor_payload
from pydantic import BaseModel
from typing import List


class Foo(BaseModel):
    count: int
    size: float = None


class Bar(BaseModel):
    apple = 'x'
    banana = 'y'


class Spam(BaseModel):
    id: str = None
    foo: Foo
    bars: List[Bar]

    class Config:
        json_encoders = {ObjectId: str}


class TestPagination:
    def test_get_pagination_cursor_payload(self):
        spam = Spam(id='lala', foo=Foo(count=1, size=1.0), bars=[Bar()])

        values = get_pagination_cursor_payload(spam, ['_id', 'id'])
        assert values[0] == 'lala'
        assert values[1] == 'lala'

        values = get_pagination_cursor_payload(spam, ['foo.count'])
        assert values[0] == 1

        values = get_pagination_cursor_payload(spam, ['bars.0.apple'])
        assert values[0] == 'x'

    def test_cursor_encoding(self):
        old_value = [ObjectId('611b158adec89d18984b7d90'), 'a', 1]
        cursor = encode_pagination_cursor(old_value)
        new_value = decode_pagination_cursor(cursor)
        assert old_value == new_value



from .errors import PaginationError
from base64 import b64decode, b64encode
from bson import ObjectId
from pydantic import BaseModel
from pydantic.generics import GenericModel
from typing import Generic, TypeVar, Any, List
import bson
import zlib

DataT = TypeVar('DataT')


class Edge(GenericModel, Generic[DataT]):
    node: DataT
    cursor: str

    class Config:
        json_encoders = {ObjectId: str}


def encode_pagination_cursor(data: List) -> str:
    byte_data = bson.BSON.encode({'v': data})
    byte_data = zlib.compress(byte_data, 9)
    return b64encode(byte_data).decode('utf-8')


def decode_pagination_cursor(data: str) -> List:
    try:
        byte_data = b64decode(data.encode('utf-8'))
        byte_data = zlib.decompress(byte_data)
        result = bson.BSON(byte_data).decode()
        return result['v']
    except Exception:
        raise PaginationError('Invalid cursor')


def get_pagination_cursor_payload(model: BaseModel, keys: List[str]) -> List[Any]:
    model_dict = model.dict()
    model_dict['_id'] = model_dict['id']

    return [
        __evaluate_dot_notation(model_dict, key) for key in keys
    ]


def __evaluate_dot_notation(data: Any, path: str):
    pieces = path.split('.')

    if len(pieces) == 1:
        return data[path]

    current_data = data

    for piece in pieces:
        if isinstance(current_data, list) or isinstance(current_data, tuple):
            current_data = current_data[int(piece)]
        else:
            current_data = current_data[piece]

    return current_data
